#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <dirent.h>
#include <unistd.h>
#include "proc_parser.h"

typedef struct {
	const char *name;
	int signal;
} SIG;

static SIG signals[] = {
	{"HUP",  SIGHUP},
	{"INT",  SIGINT},
	{"QUIT", SIGQUIT},
	{"ILL",  SIGILL},
	{"TRAP", SIGTRAP},
	{"ABRT", SIGABRT},
	{"IOT",  SIGIOT},
	{"BUS",  SIGBUS},
	{"FPE",  SIGFPE},
	{"KILL", SIGKILL},
	{"USR1", SIGUSR1},
	{"SEGV", SIGSEGV},
	{"USR2", SIGUSR2},
	{"PIPE", SIGPIPE},
	{"ALRM", SIGALRM},
	{"TERM", SIGTERM},
	{"CHLD", SIGCHLD},
	{"STOP", SIGSTOP}
};


static int parse_sig(char *arg) {
	int sig = atoi(arg);
	if (sig > 0 && sig <= NSIG)
		return sig;

	if (!strncasecmp(arg, "SIG", 3))
		arg += 3;

	for (size_t i = 0; i < sizeof(signals) / sizeof(signals[0]); i++)
		if (!strcasecmp(arg, signals[i].name))
			return signals[i].signal;

	return SIGTERM;
}

static int my_kill(pid_t pid, int signal) {
	if (kill(pid, signal)) {
		fprintf(stderr, "kill: %d: %s\n", pid, strerror(errno));
		return 1;
	}

	return 0;
}

int main(int argc, char **argv) {
	char a_flag = 0;
	char n_flag = 0;
	int signal = SIGTERM;

	int opt;
	while ((opt = getopt(argc, argv, "s:lan")) != -1) {
		switch (opt) {
			case 'l':
				for (size_t i = 0; i < sizeof(signals) / sizeof(signals[0]); i++)
					printf("%zu) %s\n", i, signals[i].name);

				return 0;

			case 's':
				signal = parse_sig(optarg);
				break;

			case 'a':
				a_flag = 1;
				break;

			case 'n':
				n_flag = 1;
				break;

			default:
				printf("kill [lsan] [pid1 pid2/str1...]\n\t-s SIG\n\t-l List all signals\n\t-a Kill all process\n\t-n STR kill processes by name\n");
				return 0;
		}
	}

	argv += optind;
	argc -= optind;

	if ((n_flag || !a_flag) && argc == 0) {
		fprintf(stderr, "kill: missing operands\n");
		return 1;
	}

	int ret = 0;

	if (a_flag || n_flag) {
		DIR *dp = opendir("/proc");
		if (dp == NULL) {
			fprintf(stderr, "kill: /proc: %s\n", strerror(errno));
			return 1;
		}

		pid_t my_pid = getpid();

		struct dirent *ep;
		while ((ep = readdir(dp)) != NULL) {
			pid_t pid = strtoul(ep->d_name, 0L, 10);
			if (pid == 0)
				continue;

			if (n_flag) {
				struct mu_proc proc;
				if (mu_proc_parser("kill", pid, &proc))
					ret = 1;

				for (int i = 0; i < argc; i++) {
					if (!strcmp(proc.prog, argv[i]))
						if (my_kill(pid, signal))
							ret = 1;
				}
			}

			else if (a_flag) {
				if (pid == my_pid)
					continue;

				if (my_kill(pid, signal))
					ret = 1;
			}
		}

		closedir(dp);
	}

	else {
		for (int i = 0; i < argc; i++)
			if (my_kill(strtoul(argv[i], 0L, 10), signal))
				ret = 1;
	}

	return ret;
}
